@@ -1,4 +1,4 @@
-;;;;  Copyright (C) 2022-2023 Andrea G. Monaco
+;;;;  Copyright (C) 2022-2024 Andrea G. Monaco
 
 ;;;;  This file is part of alisp, a lisp implementation.
 
@@ -35,6 +35,12 @@
 
 
 
+(defparameter *modules* '("this variable is deprecated per ANSI CL"))
+(defun provide (mod) "this function is deprecated per ANSI CL.  It does nothing")
+(defun require (mod &optional pl) "this function is deprecated per ANSI CL.  It does nothing")
+
+
+
 (defun describe (obj &optional str) (values))
 (defun describe-object (obj str))
 (defun inspect (obj))
@@ -66,8 +72,13 @@
 (defun identity (x) x)
 
 
+
+(defparameter *READ-DEFAULT-FLOAT-FORMAT* 'single-float)
+
+
 (defconstant PI 3.141592653589793L0)
 
+
 (defun 1+ (num)
   (+ num 1))
 
@@ -103,6 +114,39 @@
 (defun oddp (num)
   (= 1 (mod num 2)))
 
+(defun gcd (&rest nums)
+  (let ((l (length nums)))
+    (cond
+      ((= l 0)
+       0)
+      ((= l 1)
+       (abs (car nums)))
+      ((> l 2)
+       (apply 'gcd (gcd (car nums) (cadr nums)) (cddr nums)))
+      ((or (zerop (car nums)) (zerop (cadr nums)))
+       (max (abs (car nums)) (abs (cadr nums))))
+      (t
+       (gcd (min (abs (car nums)) (abs (cadr nums)))
+	    (mod (max (abs (car nums)) (abs (cadr nums)))
+		 (min (abs (car nums)) (abs (cadr nums)))))))))
+
+
+(defun lcm (&rest nums)
+  (let ((l (length nums)))
+    (cond
+      ((= l 0)
+       1)
+      ((= l 1)
+       (abs (car nums)))
+      ((> l 2)
+       (apply 'lcm (lcm (car nums) (cadr nums)) (cddr nums)))
+      ((or (zerop (car nums)) (zerop (cadr nums)))
+       0)
+      (t
+       (/ (abs (* (car nums) (cadr nums))) (gcd (car nums) (cadr nums)))))))
+
+
+
 (defun isqrt (num)
   (values (floor (sqrt num))))
 
@@ -113,6 +157,14 @@
   (complex (cos rad) (sin rad)))
 
 
+(defun upgraded-complex-part-type (type)
+  (cond
+    ((subtypep type 'integer) 'integer)
+    ((subtypep type 'ratio) 'ratio)
+    ((subtypep type 'float) 'float)))
+
+
+
 (defun logand (&rest ints)
   (lognot (apply 'logior (mapcar #'lognot ints))))
 
@@ -237,7 +289,7 @@
 
 (defun caar (list) (car (car list)))
 (defun cadr (list) (nth 1 list))
-(defun cdar (list) (cdr (car (list))))
+(defun cdar (list) (cdr (car list)))
 (defun cddr (list) (nthcdr 2 list))
 (defun caaar (list) (car (car (car list))))
 (defun caadr (list) (car (nth 1 list)))
@@ -386,6 +438,21 @@
        ,valform)))
 
 
+(defmacro prog (bins &body body)
+  `(let ,bins
+     (block nil
+       (tagbody
+	  ,@body))))
+
+
+(defmacro prog* (bins &body body)
+  `(let* ,bins
+     (block nil
+       (tagbody
+	  ,@body))))
+
+
+
 (defun every (pred &rest sequences)
   (let ((n (apply 'min (mapcar #'length sequences))))
     (dotimes (i n)
@@ -511,6 +578,46 @@
   (remove-if (complement pred) seq))
 
 
+(defun remove-duplicates (seq &key from-end test test-not (start 0) end (key #'identity))
+  (let* ((len (length seq))
+	 (n (if end (- end start) len))
+	 (removed (make-array len))
+	 leftnum
+	 (out (if from-end (reverse seq)
+		  (concatenate (if (stringp seq)
+				   'string
+				   (if (arrayp seq)
+				       'vector
+				       'list)) seq)))
+	 (tst (or test
+		  (if test-not (complement test-not) nil)
+		  #'eql)))
+    (dotimes (i n)
+      (dotimes (l (- n i 1))
+	(when (funcall tst (funcall key (elt out (+ start i)))
+		       (funcall key (elt out (+ start i l 1))))
+	  (setf (aref removed (+ start i)) t)
+	  (return nil))))
+    (setq leftnum (count-if #'not removed))
+    (dotimes (i len)
+      (dotimes (l (- len i))
+	(unless (aref removed (+ l i))
+	  (setf (elt out i) (elt out (+ l i)))
+	  (setf (aref removed (+ l i)) t)
+	  (return nil))))
+    (if (arrayp seq)
+	(adjust-array out leftnum)
+	(if (consp seq)
+	    (rplacd (nthcdr (1- leftnum) out) nil)))
+    (if from-end
+	(reverse out)
+	out)))
+
+
+(defun delete-duplicates (&rest args)
+  (apply 'remove-duplicates args))
+
+
 (defun nreverse (seq)
   (reverse seq))
 
@@ -535,6 +642,22 @@
      c))
 
 
+(defun set-difference (l1 l2 &key (key #'identity) test test-not)
+  (let ((out (concatenate 'list l1))
+	(tst (or test
+		 (if test-not (complement test-not) nil)
+		 #'eql)))
+    (dolist (l l2)
+      (setq out (remove-if (lambda (o) (funcall tst (funcall key o)
+						(funcall key l)))
+			   out)))
+    out))
+
+
+(defun nset-difference (&rest args)
+  (apply 'set-difference args))
+
+
 (defun array-rank (array)
   (length (array-dimensions array)))
 
@@ -555,6 +678,18 @@
 	  (array-dimensions array))))
 
 
+(defun upgraded-array-element-type (ts &optional env)
+  (cond
+    ((subtypep ts 'bit) 'bit)
+    ((subtypep ts 'character) 'character)
+    (t t)))
+
+
+(defun adjustable-array-p (arr)
+  t)
+
+
+
 (defun get (sym ind &optional (def nil))
   (let* ((pl (symbol-plist sym))
 	 (l (length pl)))
@@ -704,6 +839,22 @@
 
 
 
+(deftype signed-byte (&optional s)
+  (if (eq s '*)
+      'integer
+      `(integer ,(- (expt 2 (1- s))) ,(1- (expt 2 (1- s))))))
+
+
+(deftype unsigned-byte (&optional s)
+  (if (eq s '*)
+      'integer
+      `(integer 0 ,(1- (expt 2 s)))))
+
+
+(deftype mod (s)
+  `(integer 0 ,(1- s)))
+
+
 (defun consp (obj)
   (typep obj 'cons))
 
@@ -743,6 +894,9 @@
 (defun characterp (obj)
   (typep obj 'character))
 
+(defun standard-char-p (obj)
+  (typep obj 'standard-char))
+
 (defun vectorp (obj)
   (typep obj 'vector))
 
@@ -771,6 +925,11 @@
   (typep obj 'number))
 
 
+(defmacro check-type (place type &optional str)
+  `(unless (typep ,place ',type)
+     (error 'type-error)))
+
+
 
 (defun macroexpand (form)
   (do ((ret (list form t))
@@ -877,18 +1036,255 @@
   obj)
 
 
+(defmacro do-all-symbols (varres &body body)
+  (let ((packsym (gensym)))
+    `(progn
+       (dolist (,packsym (list-all-packages))
+	 (do-symbols ,(list (car varres) packsym)
+	   ,@body))
+       (let (,(car varres))
+	 ,(cadr varres)))))
+
+
+
+(defun loop-parse-accumulation (forms)
+  (let ((sym (string (car forms)))
+	var
+	out)
+    (if (string= (string (caddr forms)) "INTO")
+	(setq var (cadddr forms))
+	(setq var (gensym)))
+    (cond
+      ((or (string= sym "COLLECT")
+	   (string= sym "COLLECTING"))
+       (setq out `(setq ,var (append ,var `(,,(cadr forms))))))
+      ((or (string= sym "APPEND")
+	   (string= sym "APPENDING"))
+       (setq out `(setq ,var (append ,var ,(cadr forms)))))
+      ((or (string= sym "NCONC")
+	   (string= sym "NCONCING"))
+       (setq out `(setq ,var (nconc ,var ,(cadr forms))))))
+    (if (string= (string (caddr forms)) "INTO")
+	(values out var nil (cddddr forms))
+	(values out nil var (cddr forms)))))
+
+
+(defun loop-parse-iteration (forms)
+  (let (iter)
+    (setq iter `(,(cadr forms)))
+    (setq forms (cddr forms))
+    (do ((f forms))
+	((not f) (setq forms f))
+      (if (atom (car f))
+	  (let ((sym (string (car f))))
+	    (cond
+	      ((string= sym "FROM")
+	       (if (= (length iter) 1)
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+	       (setf (elt iter 3) (cadr f))
+	       (setq f (cddr f)))
+	      ((string= sym "DOWNFROM")
+	       (if (= (length iter) 1)
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+	       (setf (elt iter 3) (cadr f))
+	       (setf (elt iter 5) -1)
+	       (setq f (cddr f)))
+	      ((string= sym "UPFROM")
+	       (if (= (length iter) 1)
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+	       (setf (elt iter 3) (cadr f))
+	       (setf (elt iter 5) 1)
+	       (setq f (cddr f)))
+	      ((string= sym "TO")
+	       (if (= (length iter) 1)
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+	       (setf (elt iter 4) (cadr f))
+	       (setq f (cddr f)))
+	      ((string= sym "DOWNTO")
+	       (if (= (length iter) 1)
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+	       (setf (elt iter 4) (cadr f))
+	       (setf (elt iter 5) -1)
+	       (setq f (cddr f)))
+	      ((string= sym "UPTO")
+               (if (= (length iter) 1)
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+	       (setf (elt iter 4) (cadr f))
+	       (setf (elt iter 5) 1)
+	       (setq f (cddr f)))
+	      ((string= sym "BELOW")
+	       (if (= (length iter) 1)
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+	       (setf (elt iter 4) (1- (cadr f)))
+	       (setq f (cddr f)))
+	      ((string= sym "ABOVE")
+	       (if (= (length iter) 1)
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+	       (setf (elt iter 4) (1+ (cadr f)))
+	       (setq f (cddr f)))
+	      ((string= sym "BY")
+	       (if (= (length iter) 1)
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+	       (cond
+		 ((eq (car iter) :from)
+		  (setf (elt iter 6) (cadr f)))
+		 ((eq (car iter) :in)
+		  (setf (elt iter 4) (cadr f)))
+		 ((eq (car iter) :on)
+		  (setf (elt iter 3) (cadr f))))
+	       (setq f (cddr f)))
+	      ((string= sym "IN")
+	       (setq iter `(:in ,(car iter) nil ,(cadr f) #'cdr))
+	       (setq f (cddr f)))
+	      ((string= sym "ON")
+	       (setq iter `(:on ,(car iter) ,(cadr f) #'cdr))
+	       (setq f (cddr f)))
+	      ((string= sym "=")
+	       (setq iter `(:eq ,(car iter) ,(cadr f) nil))
+	       (setq f (cddr f)))
+	      ((string= sym "THEN")
+	       (setf (elt iter 3) (cadr f))
+	       (setq f (cddr f)))
+	      ((string= sym "BEING")
+	       (setq iter `(:hashtable ,(car iter) nil nil nil))
+	       (setq f (cdr f))
+	       (if (or (string= (string (car f)) "EACH")
+		       (string= (string (car f)) "THE"))
+		   (setq f (cdr f)))
+	       (when (or (string= (string (car f)) "HASH-VALUE")
+			 (string= (string (car f)) "HASH-VALUES"))
+		 (setf (elt iter 2) (elt iter 1))
+		 (setf (elt iter 1) nil)
+		 (setq f (cdr f)))
+	       (if (or (string= (string (car f)) "HASH-KEY")
+		       (string= (string (car f)) "HASH-KEYS"))
+		   (setq f (cdr f)))
+	       (when (or (string= (string (car f)) "IN")
+			 (string= (string (car f)) "OF"))
+		 (setf (elt iter 4) (cadr f))
+		 (setq f (cddr f)))
+	       (when (string= (string (car f)) "USING")
+		 (setf f (cdr f))
+		 (if (string= (string (caar f)) "HASH-VALUE")
+		     (setf (elt iter 2) (cadar f))
+		     (setf (elt iter 1) (cadar f)))
+		 (setq f (cdr f))))
+	      (t
+	       (setq forms f)
+	       (return nil))))
+	  (progn
+	    (setq forms f)
+	    (return nil))))
+    (if (member (car iter) '(:from :in))
+	(setf (elt iter 2) (gensym)))
+    (when (eq (car iter) :hashtable)
+      (unless (elt iter 1)
+	(setf (elt iter 1) (gensym)))
+      (unless (elt iter 2)
+	(setf (elt iter 2) (gensym)))
+      (setf (elt iter 3) (gensym)))
+    (values iter forms)))
+
+
+(defun loop-parse-conditional (forms loopname)
+  (let (out
+	docl
+	elsecl
+	retst
+	vars
+	returnvar
+	(sym (string (car forms))))
+    (cond
+      ((or (string= sym "IF")
+	   (string= sym "WHEN"))
+       (setq out `(if ,(cadr forms) nil))
+       (setq forms (cddr forms)))
+      ((string= sym "UNLESS")
+       (setq out `(if (not ,(cadr forms)) nil))
+       (setq forms (cddr forms))))
+    (do ((f forms))
+	((not f) (setq forms f))
+      (if (atom (car f))
+	  (let ((sym (string (car f))))
+	    (cond
+	      ((or (string= sym "IF")
+		   (string= sym "WHEN")
+		   (string= sym "UNLESS"))
+	       (multiple-value-bind (res frm vs rv)
+		   (loop-parse-conditional f loopname)
+		 (if (= (length out) 3)
+		     (setq docl (cons res docl))
+		     (setq elsecl (cons res elsecl)))
+		 (setq vars (append vs vars))
+		 (if rv
+		     (setq returnvar rv))
+		 (setq f frm)))
+	      ((string= sym "ELSE")
+	       (setq out (append out (list nil)))
+	       (setq f (cdr f)))
+	      ((or (string= sym "DO")
+		   (string= sym "DOING")
+		   (string= sym "AND"))
+	       (setq f (cdr f)))
+	      ((string= sym "RETURN")
+	       (setq retst `(return-from ,loopname ,(cadr f)))
+	       (if (= (length out) 3)
+		   (setq docl (cons retst docl))
+		   (setq elsecl (cons retst elsecl)))
+	       (setq f (cddr f)))
+	      ((or (string= sym "COLLECT")
+		   (string= sym "COLLECTING")
+		   (string= sym "APPEND")
+		   (string= sym "APPENDING")
+		   (string= sym "NCONC")
+		   (string= sym "NCONCING"))
+	       (multiple-value-bind (res var1 var2 frm) (loop-parse-accumulation f)
+		 (if (= (length out) 3)
+		     (setq docl (cons res docl))
+		     (setq elsecl (cons res elsecl)))
+		 (setq vars (cons (list (or var1 var2)) vars))
+		 (if var2
+		     (setq returnvar var2))
+		 (setq f frm)))
+	      ((string= sym "END")
+	       (setq forms (cdr f))
+	       (return nil))
+	      (t
+	       (setq forms f)
+	       (return nil))))
+	  (progn
+	    (if (= (length out) 3)
+		(setq docl (cons (car f) docl))
+		(setq elsecl (cons (car f) elsecl)))
+	    (setq f (cdr f)))))
+    (setq docl (reverse docl))
+    (setq docl (cons 'progn docl))
+    (setf (elt out 2) docl)
+    (when (= (length out) 4)
+      (setq elsecl (reverse elsecl))
+      (setq elsecl (cons 'progn elsecl))
+      (setf (elt out 3) elsecl))
+    (values out forms vars returnvar)))
+
+
 (defmacro loop (&body forms)
   (let (block-name
 	(tagname (gensym))
+	(middle-block-name (gensym))
 	(inner-block-name (gensym))
 	initially-forms
 	finally-forms
 	do-forms
-	vars var)
+	iters
+	var vars parvars
+	lets l
+	outdestbinds od
+	indestbinds id
+	returnvar)
     (do ()
 	((not forms))
       (let ((form (car forms)))
-	(if (atom form)
+	(if (symbolp form)
 	    (let ((sym (string form)))
 	      (cond
                 ((string= sym "INITIALLY")
@@ -899,71 +1295,192 @@
 		 (do ((f (cdr forms) (cdr f)))
 		     ((atom (car f)) (setq forms f))
 		   (setq finally-forms (append finally-forms `(,(car f))))))
+		((string= sym "WITH")
+		 (when parvars
+		   (setq vars (cons parvars vars))
+		   (setq parvars nil))
+		 (setq var `(,(cadr forms) ,(cadddr forms)))
+		 (setq parvars (cons var parvars))
+		 (setq forms (cddddr forms)))
+		((string= sym "AND")
+		 (setq var `(,(cadr forms) ,(cadddr forms)))
+		 (setq parvars (cons var parvars))
+		 (setq forms (cddddr forms)))
+		((or (string= sym "IF")
+		     (string= sym "WHEN")
+		     (string= sym "UNLESS"))
+		 (multiple-value-bind (res frm vs rv)
+		     (loop-parse-conditional forms block-name)
+		   (setq do-forms (append do-forms `(,res)))
+		   (setq vars (append vs vars))
+		   (if rv
+		       (setq returnvar rv))
+		   (setq forms frm)))
 		((or (string= sym "DO")
 		     (string= sym "DOING"))
-		 (do ((f (cdr forms) (cdr f)))
-		     ((atom (car f)) (setq forms f))
-		   (setq do-forms (append do-forms `(,(car f))))))
+		 (setq forms (cdr forms)))
+		((string= sym "RETURN")
+		 (setq do-forms (append do-forms `((return-from ,block-name ,(cadr forms)))))
+		 (setq forms (cddr forms)))
 		((string= sym "NAMED")
 		 (setq block-name (cadr forms))
 		 (setq forms (cddr forms)))
 		((or (string= sym "FOR")
 		     (string= sym "AS"))
-		 (setq var `(,(cadr forms) 0 nil 1 1))
-		 (setq vars (append vars `(,var)))
-		 (setq forms (cddr forms)))
-		((string= sym "FROM")
-		 (setf (elt var 1) (cadr forms))
-		 (setq forms (cddr forms)))
-		((or (string= sym "DOWNFROM"))
-		 (setf (elt var 1) (cadr forms))
-		 (setf (elt var 3) -1)
-		 (setq forms (cddr forms)))
-		((or (string= sym "UPFROM"))
-		 (setf (elt var 1) (cadr forms))
-		 (setf (elt var 3) 1)
-		 (setq forms (cddr forms)))
-		((or (string= sym "TO"))
-		 (setf (elt var 2) (cadr forms))
-		 (setq forms (cddr forms)))
-		((or (string= sym "DOWNTO"))
-		 (setf (elt var 2) (cadr forms))
-		 (setf (elt var 3) -1)
-		 (setq forms (cddr forms)))
-		((or (string= sym "UPTO"))
-		 (setf (elt var 2) (cadr forms))
-		 (setf (elt var 3) 1)
-		 (setq forms (cddr forms)))
-		((or (string= sym "BELOW"))
-		 (setf (elt var 2) (1- (cadr forms)))
-		 (setq forms (cddr forms)))
-		((or (string= sym "ABOVE"))
-		 (setf (elt var 2) (1+ (cadr forms)))
-		 (setq forms (cddr forms)))
-		((or (string= sym "BY"))
-		 (setf (elt var 4) (cadr forms))
-		 (setq forms (cddr forms)))))
+		 (multiple-value-bind (res frm) (loop-parse-iteration forms)
+		   (setq iters (append iters `(,res)))
+		   (setq forms frm)))
+		((or (string= sym "COLLECT")
+		     (string= sym "COLLECTING")
+		     (string= sym "APPEND")
+		     (string= sym "APPENDING")
+		     (string= sym "NCONC")
+		     (string= sym "NCONCING"))
+		 (multiple-value-bind (res var1 var2 frm) (loop-parse-accumulation forms)
+		   (setq do-forms (append do-forms `(,res)))
+		   (setq vars (cons (list (or var1 var2)) vars))
+		   (if var2
+		       (setq returnvar var2))
+		   (setq forms frm)))
+		(t
+		 (setq do-forms (append do-forms `(,form)))
+		 (setq forms (cdr forms)))))
 	    (progn
 	      (setq do-forms (append do-forms `(,form)))
 	      (setq forms (cdr forms))))))
+    (if parvars
+	(setq vars (cons parvars vars)))
+    (setq vars (reverse vars))
+    (dolist (v vars)
+      (if lets
+	  (progn
+	    (setf (car l) `(let ,v nil))
+	    (setq l (cddar l)))
+	  (progn
+	    (setq lets `(let ,v nil))
+	    (setq l (cddr lets)))))
+    (dolist (i iters)
+      (let ((outdb (cond
+		     ((eq (car i) :hashtable)
+		      `(let (,(cadddr i))
+			 (maphash (lambda (k v) (setq ,(elt i 3) (cons (cons k v) ,(elt i 3)))) ,(elt i 4))
+			 nil))
+		     ((member (car i) '(:on :eq))
+		      `(destructuring-bind (,(cadr i))
+			   (list ,(caddr i))
+			 nil))
+		     (t
+		      `(destructuring-bind (,(caddr i))
+			   (list ,(cadddr i))
+			 nil))))
+	    (listcheck `(progn
+			  (if (endp ,(elt i 3))
+			      (return-from ,middle-block-name nil))
+			  nil))
+	    (middb (if (eq (car i) :hashtable)
+		       `(let ((,(cadr i) (caar ,(elt i 3)))
+			      (,(caddr i) (cdar ,(elt i 3))))
+			  nil nil)
+		       `(destructuring-bind (,(cadr i))
+			    ,(if (eq (car i) :from)
+				 `(list ,(cadddr i))
+				 `(list (car ,(cadddr i))))
+			  nil)))
+	    (indb (if (eq (car i) :hashtable)
+		      `(let ((,(caddr i) (cdar ,(elt i 3)))
+			     (,(cadr i) (caar ,(elt i 3))))
+			 nil
+			 nil)
+		      `(destructuring-bind (,(cadr i))
+			   ,(if (eq (car i) :from)
+				`(list ,(caddr i))
+				`(list (car ,(caddr i))))
+			 nil))))
+	(if outdestbinds
+	    (progn
+	      (setf (car od) outdb)
+	      (setq od (cdddar od))
+	      (when (member (car i) '(:from :in :hashtable))
+	      	(when (not (eq (car i) :from))
+		  (setf (car od) listcheck)
+		  (setq od (cddar od)))
+		(setf (car od) middb)
+		(setq od (cdddar od))
+		(setf (car id) indb)
+		(setq id (cdddar id))))
+	    (progn
+	      (setf outdestbinds outdb)
+	      (setq od (cdddr outdestbinds))
+	      (when (member (car i) '(:from :in :hashtable))
+		(when (not (eq (car i) :from))
+		  (setf (car od) listcheck)
+		  (setq od (cddar od)))
+		(setf (car od) middb)
+		(setq od (cdddar od))
+		(setf indestbinds indb)
+		(setq id (cdddr indestbinds)))))))
     `(block ,block-name
-       (let ,(mapcar (lambda (x) `(,(car x) ,(cadr x))) vars)
-	 ,@initially-forms
-	 (block ,inner-block-name
-	   (tagbody
-	      ,tagname
-	      (if (or ,@(mapcar (lambda (x) (if (elt x 2)
-						(if (= (elt x 3) 1)
-						    `(> ,(elt x 0) ,(elt x 2))
-						    `(< ,(elt x 0) ,(elt x 2)))
-						nil)) vars))
-		  (return-from ,inner-block-name nil))
-	      ,@do-forms
+       ,(let ((inner-body
+	       `(progn
+		  (block ,middle-block-name
+		    ,(let ((pre-initially-forms-body
+			    `(progn
+			       ,@initially-forms
+			       (block ,inner-block-name
+				 (tagbody
+				    ,tagname
+				    (if (or ,@(mapcar (lambda (x) (cond
+								    ((eq (car x) :from)
+								     (if (elt x 4)
+									 (if (= (elt x 5) 1)
+									     `(> ,(elt x 2) ,(elt x 4))
+									     `(< ,(elt x 2) ,(elt x 4)))))
+								    ((eq (car x) :on)
+								     `(endp ,(elt x 1)))
+								    ((eq (car x) :in)
+								     `(endp ,(elt x 2)))
+								    ((eq (car x) :hashtable)
+								     `(endp ,(elt x 3)))))
+						      iters))
+					,(if returnvar
+					     `(return-from ,block-name ,returnvar)
+					     `(return-from ,inner-block-name nil)))
+				    ,(let ((post-initially-forms-body
+					    `(progn
+					       ,@do-forms
+					       ,@(mapcar (lambda (x) (cond
+								       ((eq (car x) :on)
+									`(setq ,(elt x 1) (funcall ,(elt x 3) ,(elt x 1))))
+								       ((eq (car x) :eq)
+									`(setq ,(elt x 1) ,(or (elt x 3) (elt x 2))))
+								       ((eq (car x) :from)
+									`(setq ,(elt x 2) (+ ,(elt x 1) ,(* (elt x 5) (elt x 6)))))
+								       ((eq (car x) :in)
+									`(setq ,(elt x 2) (funcall ,(elt x 4) ,(elt x 2))))
+								       ((eq (car x) :hashtable)
+									`(setq ,(elt x 3) (cdr ,(elt x 3))))))
+							 iters)
+					       (go ,tagname))))
+				       (if indestbinds
+					   (progn
+					     (setf (car id) post-initially-forms-body)
+					     indestbinds)
+					   post-initially-forms-body))))
+			       ,@finally-forms
+			       (return-from ,block-name nil))))
+		       (if outdestbinds
+			   (progn
+			     (setf (car od) pre-initially-forms-body)
+			     outdestbinds)
+			   pre-initially-forms-body)))
+		  ,@finally-forms)))
+	  (if lets
 	      (progn
-		,@(mapcar (lambda (x) `(setq ,(elt x 0) (+ ,(elt x 0) ,(* (elt x 3) (elt x 4))))) vars))
-	      (go ,tagname)))
-	 ,@finally-forms
-	 nil))))
+		(setf (car l) inner-body)
+		lets)
+	      inner-body))
+       nil)))
+
 
 
 (defun format (out fstr &rest args)
@@ -1001,34 +1518,38 @@
 
 
 (export '(*features* machine-instance machine-type machine-version
-	  short-site-name long-site-name describe describe-object inspect
-	  lambda-parameters-limit call-arguments-limit multiple-values-limit
-	  array-rank-limit array-dimension-limit array-total-size-limit
-	  char-code-limit lambda-list-keywords identity pi 1+ 1- minusp plusp
-	  abs zerop signum mod rem evenp oddp isqrt conjugate cis logand
-	  logandc1 logandc2 logeqv lognand lognor logorc1 logorc2 logxor boole-1
-	  boole-2 boole-andc1 boole-andc2 boole-and boole-c1 boole-c2 boole-clr
-	  boole-eqv boole-ior boole-nand boole-nor boole-orc1 boole-orc2
-	  boole-set boole-xor boole *gensym-counter* gensym gentemp first second
-	  third fourth fifth sixth seventh eighth ninth tenth rest caar cadr
-	  cdar cddr caaar caadr cadar caddr cdaar cdadr cddar cdddr caaaar
-	  caaadr caadar caaddr cadaar cadadr caddar cadddr cdaaar cdaadr cdadar
-	  cdaddr cddaar cddadr cdddar cddddr make-list endp butlast nbutlast
-	  acons pairlis when unless incf decf and or cond otherwise case return
-	  multiple-value-bind every some notany notevery member member-if
-	  member-if-not find find-if find-if-not assoc assoc-if assoc-if-not
-	  position position-if position-if-not count count-if count-if-not
-	  remove remove-if-not delete delete-if delete-if-not nreverse adjoin
-	  fill push pop array-rank array-dimension array-total-size
-	  array-in-bounds-p get get-properties char string/= char-equal
+	  short-site-name long-site-name *modules* provide require describe
+	  describe-object inspect lambda-parameters-limit call-arguments-limit
+	  multiple-values-limit array-rank-limit array-dimension-limit
+	  array-total-size-limit char-code-limit lambda-list-keywords identity
+	  *read-default-float-format* pi 1+ 1- minusp plusp abs zerop signum mod
+	  rem evenp oddp gcd lcm isqrt conjugate cis upgraded-complex-part-type
+	  logand logandc1 logandc2 logeqv lognand lognor logorc1 logorc2 logxor
+	  boole-1 boole-2 boole-andc1 boole-andc2 boole-and boole-c1 boole-c2
+	  boole-clr boole-eqv boole-ior boole-nand boole-nor boole-orc1
+	  boole-orc2 boole-set boole-xor boole *gensym-counter* gensym gentemp
+	  first second third fourth fifth sixth seventh eighth ninth tenth rest
+	  caar cadr cdar cddr caaar caadr cadar caddr cdaar cdadr cddar cdddr
+	  caaaar caaadr caadar caaddr cadaar cadadr caddar cadddr cdaaar cdaadr
+	  cdadar cdaddr cddaar cddadr cdddar cddddr make-list endp butlast
+	  nbutlast acons pairlis when unless incf decf and or cond otherwise
+	  case return multiple-value-bind prog prog* every some notany notevery
+	  member member-if member-if-not find find-if find-if-not assoc assoc-if
+	  assoc-if-not position position-if position-if-not count count-if
+	  count-if-not remove remove-if-not delete delete-if delete-if-not
+	  remove-duplicates delete-duplicates nreverse adjoin fill push pop
+	  set-difference nset-difference array-rank array-dimension
+	  array-total-size array-in-bounds-p upgraded-array-element-type
+	  adjustable-array-p get get-properties char string/= char-equal
 	  digit-char digit-char-p char-int string-upcase string-downcase
 	  string-capitalize nstring-upcase nstring-downcase nstring-capitalize
-	  string-left-trim string-right-trim string-trim defpackage consp listp
-	  symbolp keywordp functionp packagep integerp rationalp floatp complexp
-	  random-state-p characterp vectorp arrayp sequencep stringp
-	  hash-table-p pathnamep streamp realp numberp macroexpand equal equalp
-	  fdefinition complement mapc terpri write-line write-sequence prin1
-	  princ print loop format encode-universal-time))
+	  string-left-trim string-right-trim string-trim defpackage signed-byte
+	  unsigned-byte consp listp symbolp keywordp functionp packagep integerp
+	  rationalp floatp complexp random-state-p characterp standard-char-p
+	  vectorp arrayp sequencep stringp hash-table-p pathnamep streamp realp
+	  numberp check-type macroexpand equal equalp fdefinition complement
+	  mapc terpri write-line write-sequence prin1 princ print do-all-symbols
+	  loop format encode-universal-time))
 
 
 
