@@ -1,7 +1,7 @@
 #!/usr/bin/perl -w
 use strict;
 
-# Copyright (C) 2022-2023 Andrea G. Monaco
+# Copyright (C) 2022-2024 Andrea G. Monaco
 # 
 # This file is part of alisp, a lisp implementation.
 #
@@ -132,10 +132,15 @@ make_test ("`(,@() ,.())", "NIL");
 make_test ("`(1 2 ,@() ,.() 3)", "(1 2 3)");
 make_test ("``(1 2 ,,@() ,,.())", "`(1 2)");
 make_test ("``(,@())", "`(,\@NIL)");
+make_test ("`(\n,@(\n))", "NIL");
 make_test ("`(,\@5)", "5");
 make_test ("`(1 ,\@5)", "(1 . 5)");
 make_test ("`(,\@'(3 4))", "(3 4)");
 make_test ("`(1 2 ,\@'(3 . 4))", "(1 2 3 . 4)");
+make_test ("'(`(\n,@(2)))", "(`(,@(2)))");
+make_test ("'(0 . `(\n,@(2)))", "(0 . `(,@(2)))");
+make_test ("``(,,@(list 0 1 2))", "`(,0 ,1 ,2)");
+make_test ("``(,,.(list 0 1 2))", "`(,0 ,1 ,2)");
 make_test ("`(,@``(3 4))", "`(3 4)");
 make_test ("`(,1 ,@(cdr '(1 2 3 4)))", "(1 2 3 4)");
 make_test ("`(,1 ,@(cdr '(1 2 3 4)) ,5)", "(1 2 3 4 5)");
@@ -176,6 +181,18 @@ make_test ("#(\n" .
 	   ")\n" .
 	   ")", "#(NIL)");
 make_test ("'#:a", "#:A");
+make_test ("'#:car", "#:CAR");
+make_test ("#b01010", "10");
+make_test ("#xf4/f", "244/15");
+make_test ("#o10", "8");
+make_test ("#29rh", "17");
+make_test ("(setq *read-base* 16)", "16");
+make_test ("(+ a f)", "25");
+make_test ("(setq *read-base* 8)", "8");
+make_test ("10", "8");
+make_test ("10/12", "4/5");
+make_test ("(typep '9 'symbol)", "T");
+make_test ("(setq *read-base* 12)", "10");
 
 
 # eval tests
@@ -192,6 +209,7 @@ make_test ("(multiple-value-list (values))", "NIL");
 make_test ("(multiple-value-list (values 1 2 3))", "(1 2 3)");
 make_test ("(multiple-value-call (lambda ()))", "NIL");
 make_test ("(multiple-value-call #'+ (floor 5 2) (values) 1)", "4");
+make_test ("(multiple-value-call (lambda (x) x) 'i)", "I");
 make_test ("(if nil 1)", "NIL");
 make_test ("(common-lisp:if '(1) 2)", "2");
 make_test ("(cl:eq 1 1)", "NIL");
@@ -237,9 +255,13 @@ make_test ("(let* ((x 0)) (declare (special x)) x)", "0");
 make_test ("(let* ((x 0)) (declare (special x)) (let ((x 1)) x))", "1");
 make_test ("(let* ((x 0)) (declare (special x)) (let ((x 1)) (eval 'x)))", "0");
 make_test ("(let* ((a)) a)", "NIL");
+make_test ("(defun fooo () (list x y z))", "FOOO");
+make_test ("(progv '(x y z) '(0 1 2) (fooo))", "(0 1 2)");
 make_test ("(flet ((a () (write \"\"))) (a))", "\"\"\n\"\"");
+make_test ("(let ((i 0)) (flet ((fun () (setq i 1))) (fun)) i)", "1");
 make_test ("(labels ((a NIL (write \"\"))) (a))", "\"\"\n\"\"");
 make_test ("(labels ((a () (b)) (b () (write \"\"))) (a))", "\"\"\n\"\"");
+make_test ("(let ((i 0)) (labels ((fun () (setq i 1))) (fun)) i)", "1");
 make_test ("(macrolet ((w (a) `(write ,a))) (w \"\"))", "\"\"\n\"\"");
 make_test ("(eval-when (:compile-toplevel compile) 0)", "NIL");
 make_test ("(eval-when (:load-toplevel load) 0)", "NIL");
@@ -260,6 +282,7 @@ make_test ("(setf c (+ c 1))", "5");
 make_test ("c", "5");
 make_test ("(setq)", "NIL");
 make_test ("(setq b 9 c 4 c (+ c 1))", "5");
+make_test ("(let ((f 0)) (setq g 1) f)", "0");
 make_test ("(psetq)", "NIL");
 make_test ("(psetq xx 0)", "NIL");
 make_test ("(psetq xx 1 yy xx)", "NIL");
@@ -270,6 +293,7 @@ make_test ("(setf (car tt) 'f)", "F");
 make_test ("tt", "(F B D)");
 make_test ("(setf (cdr tt) 'g)", "G");
 make_test ("tt", "(F . G)");
+make_test ("(let ((l '(1 2))) (setf (nth 1 l) 3) l)", "(1 3)");
 make_test ("(defparameter str \"abcdef\")", "STR");
 make_test ("(setf (aref str 2) #\\z)", "#\\z");
 make_test ("str", "\"abzdef\"");
@@ -356,6 +380,8 @@ make_test ("(inc)", "2");
 make_test ("(inc)", "3");
 make_test ("(defmacro f () (f))", "F");
 make_test ("(defmacro f () (f))", "F");
+make_test ("(defmacro mac (x (y (z))) `(list ,x ,y ,z))", "MAC");
+make_test ("(mac 0 (1 (2)))", "(0 1 2)");
 make_test ("(load \"cl.lisp\")", "T");
 make_test ("(open-stream-p *standard-input*)", "T");
 make_test ("(input-stream-p *standard-input*)", "T");
@@ -393,10 +419,30 @@ make_test ("(typep 0.1 'real)", "T");
 make_test ("(typep 0 'rational)", "T");
 make_test ("(typep 0.1 'rational)", "NIL");
 make_test ("(typep 0 'integer)", "T");
+make_test ("(typep 2/1 'integer)", "T");
+make_test ("(typep (+ 1/2 1/2) 'integer)", "T");
+make_test ("(typep 0 '(integer 0 0))", "T");
+make_test ("(typep 0 '(integer 1 2))", "NIL");
+make_test ("(typep 0 '(integer 1))", "NIL");
+make_test ("(typep 0 '(integer 1 *))", "NIL");
 make_test ("(typep 0 'bignum)", "T");
 make_test ("(typep 0 'fixnum)", "NIL");
+make_test ("(typep 0 'bit)", "T");
+make_test ("(typep 1 'bit)", "T");
+make_test ("(typep 2 'bit)", "NIL");
+make_test ("(typep 0 'signed-byte)", "T");
+make_test ("(typep 3 '(signed-byte 3))", "T");
+make_test ("(typep 4 '(signed-byte 3))", "NIL");
+make_test ("(typep -4 '(signed-byte 3))", "T");
+make_test ("(typep -5 '(signed-byte 3))", "NIL");
+make_test ("(typep 10 '(unsigned-byte))", "T");
+make_test ("(typep 7 '(unsigned-byte 3))", "T");
+make_test ("(typep 8 '(unsigned-byte 3))", "NIL");
+make_test ("(typep -1 '(mod 2))", "NIL");
+make_test ("(typep 1 '(mod 2))", "T");
+make_test ("(typep 2 '(mod 2))", "NIL");
+make_test ("(typep 2/1 'ratio)", "NIL");
 make_test ("(typep 1/2 'ratio)", "T");
-make_test ("(typep 3/1 'ratio)", "NIL");
 make_test ("(typep 0.1 'float)", "T");
 make_test ("(typep 0.1 'single-float)", "T");
 make_test ("(typep 0.1 'double-float)", "T");
@@ -405,6 +451,11 @@ make_test ("(typep 0.1 'long-float)", "T");
 make_test ("(typep 1.0 'complex)", "NIL");
 make_test ("(typep (complex 1 1) 'complex)", "T");
 make_test ("(typep (make-random-state) 'random-state)", "T");
+make_test ("(typep #\\a 'character)", "T");
+make_test ("(typep #\\ò 'character)", "T");
+make_test ("(typep #\\ò 'base-char)", "T");
+make_test ("(typep #\\ò 'standard-char)", "NIL");
+make_test ("(typep #\\! 'standard-char)", "T");
 make_test ("(typep \"abc\" 'string)", "T");
 make_test ("(typep \"abc\" '(string))", "T");
 make_test ("(typep \"abc\" 'array)", "T");
@@ -450,9 +501,22 @@ make_test ("(subtypep 'integer 'cons)", "NIL\nT");
 make_test ("(subtypep 'short-float 'long-float)", "T\nT");
 make_test ("(subtypep 'long-float 'short-float)", "T\nT");
 make_test ("(subtypep 'long-float 'single-float)", "T\nT");
+make_test ("(subtypep 'float 'number)", "T\nT");
+make_test ("(subtypep 'float 'real)", "T\nT");
+make_test ("(subtypep 'character 'float)", "NIL\nT");
+make_test ("(subtypep 'character 'integer)", "NIL\nT");
+make_test ("(subtypep 'character 'bit)", "NIL\nT");
 make_test ("(deftype not-integer () `(not integer))", "NOT-INTEGER");
 make_test ("(typep 0 'not-integer)", "NIL");
 make_test ("(typep \"\" 'not-integer)", "T");
+make_test ("(deftype this-object (obj) `(eql ,obj))", "THIS-OBJECT");
+make_test ("(typep 0 '(this-object 0))", "T");
+make_test ("(typep 0 '(this-object 1))", "NIL");
+make_test ("(typep \"a\" '(eql \"a\"))", "NIL");
+make_test ("(typep 1 '(member 1 2 3))", "T");
+make_test ("(typep 4 '(member 1 2 3))", "NIL");
+make_test ("(typep 4 '(member))", "NIL");
+make_test ("(let ((i 0)) (check-type i integer))", "NIL");
 make_test ("(define-setf-expander foo ())", "FOO");
 make_test ("(get-setf-expansion '(foo))", "NIL");
 make_test ("(defun (setf foo2) (a1 a2) (list a1 a2))", "(SETF FOO2)");
@@ -469,6 +533,7 @@ make_test ("(intern \"hi\" 'keyword)", ":|hi|\n:EXTERNAL");
 make_test ("(find-symbol \"CAR\")", "CAR\n:INHERITED");
 make_test ("(find-symbol \"CAR\" 'cl)", "CAR\n:EXTERNAL");
 make_test ("(find-symbol \"car\")", "NIL\nNIL");
+make_test ("(find-symbol \"INTERNAL\" 'keyword)", ":INTERNAL\n:EXTERNAL");
 make_test ("(unintern :hi)", "NIL");
 make_test ("(unintern '|hi|)", "T");
 make_test ("(intern \"hi\")", "|hi|\nNIL");
@@ -581,6 +646,7 @@ make_test ("(car '(1 2))", "1");
 make_test ("(cdr '(1 2))", "(2)");
 make_test ("(car ())", "NIL");
 make_test ("(cdr ())", "NIL");
+make_test ("(cdar '((0 . 1)))", "1");
 make_test ("(setf tt (list 'a 'b 'd))", "(A B D)");
 make_test ("(rplaca tt 'f)", "(F B D)");
 make_test ("tt", "(F B D)");
@@ -588,6 +654,7 @@ make_test ("(rplacd tt 'g)", "(F . G)");
 make_test ("tt", "(F . G)");
 make_test ("(nth 0 '(0))", "0");
 make_test ("(nth 1 '(0))", "NIL");
+make_test ("(nth 3 nil)", "NIL");
 make_test ("(nthcdr 1 ())", "NIL");
 make_test ("(nthcdr 2 '(a b c))", "(C)");
 make_test ("(nthcdr 1 '(0 . 1))", "1");
@@ -619,11 +686,19 @@ make_test ("(length #(1 2))", "2");
 make_test ("(length '(0 1 2))", "3");
 make_test ("(list-length '(0 1 2 3))", "4");
 make_test ("(make-array 4)", "#(NIL NIL NIL NIL)");
-make_test ("(make-array '(1 2 3))", "#<ARRAY, RANK 3>");
-make_test ("(make-array nil)", "#<ARRAY, RANK 0>");
+make_test ("(make-array '(2 3))", "#2A((NIL NIL NIL) (NIL NIL NIL))");
+make_test ("(let ((*print-array* nil)) (write (make-array '(1 2 3))))", "#<ARRAY, RANK 3>\n#3A(((NIL NIL NIL) (NIL NIL NIL)))");
+make_test ("(make-array '(1 2 3))", "#3A(((NIL NIL NIL) (NIL NIL NIL)))");
+make_test ("(make-array nil)", "#0ANIL");
 make_test ("(vector)", "#()");
 make_test ("(vector 'a 'b \"\")", "#(A B \"\")");
+make_test ("#*", "#*");
+make_test ("#0*", "#*");
+make_test ("#*011001", "#*011001");
+make_test ("#4*010", "#*0100");
+make_test ("#4*011", "#*0111");
 make_test ("(array-has-fill-pointer-p \"aaa\")", "NIL");
+make_test ("(length #*011001)", "6");
 make_test ("(array-rank #())", "1");
 make_test ("(array-rank (make-array nil))", "0");
 make_test ("(array-rank (make-array '(1 2)))", "2");
@@ -642,16 +717,45 @@ make_test ("(array-row-major-index (make-array '(1 2 3)) 0 0 0)", "0");
 make_test ("(array-row-major-index (make-array '(2 3 4)) 0 0 1)", "1");
 make_test ("(array-row-major-index (make-array '(2 3 4)) 0 2 0)", "8");
 make_test ("(array-row-major-index (make-array '(2 3 4)) 1 0 0)", "12");
+make_test ("(let ((arr (make-array 5))) (adjust-array arr 3) arr)", "#(NIL NIL NIL)");
+make_test ("(let ((arr (make-array 2))) (adjust-array arr 3) arr)", "#(NIL NIL NIL)");
+make_test ("(let ((str \"abcdef\")) (adjust-array str 4) str)", "\"abcd\"");
+make_test ("(let ((str \"abcdef\")) (adjust-array str 8) (length str))", "8");
+make_test ("(adjust-array #(0 1 2 4 5 3) 5)", "#(0 1 2 4 5)");
 make_test ("(defparameter tbl (make-hash-table))", "TBL");
-make_test ("tbl", "#<HASH-TABLE EQ 0/1024>");
+make_test ("tbl", "#<HASH-TABLE EQL 0/1024>");
 make_test ("(hash-table-size tbl)", "1024");
 make_test ("(hash-table-count tbl)", "0");
-make_test ("(hash-table-test tbl)", "EQ");
+make_test ("(hash-table-test tbl)", "EQL");
+make_test ("(gethash 10 tbl)", "NIL\nNIL");
+make_test ("(setf (gethash 10 tbl) 11)", "11");
+make_test ("(gethash 10 tbl)", "11\nT");
+make_test ("(hash-table-size tbl)", "1024");
+make_test ("(hash-table-count tbl)", "1");
+make_test ("(setf (gethash #\\a tbl) 12)", "12");
+make_test ("(maphash (lambda (k v) (write k) (write v)) tbl)", "1011#\\a12\nNIL");
+make_test ("(hash-table-count tbl)", "2");
+make_test ("(defparameter tbl (make-hash-table :test 'eq))", "TBL");
+make_test ("tbl", "#<HASH-TABLE EQ 0/1024>");
 make_test ("(gethash 'k tbl)", "NIL\nNIL");
 make_test ("(setf (gethash 'k tbl) 10)", "10");
+make_test ("(hash-table-size tbl)", "1024");
+make_test ("(hash-table-count tbl)", "1");
 make_test ("(gethash 'k tbl)", "10\nT");
 make_test ("(remhash 'k tbl)", "T");
 make_test ("(gethash 'k tbl)", "NIL\nNIL");
+make_test ("(hash-table-count tbl)", "0");
+make_test ("(defparameter tbl (make-hash-table :test 'equal))", "TBL");
+make_test ("(setf (gethash \"hello\" tbl) 12)", "12");
+make_test ("(gethash \"hello\" tbl)", "12\nT");
+make_test ("(setf (gethash \"helloo\" tbl) 13)", "13");
+make_test ("(gethash \"helloo\" tbl)", "13\nT");
+make_test ("(hash-table-count tbl)", "2");
+make_test ("(remhash \"hello\" tbl)", "T");
+make_test ("(gethash \"hello\" tbl)", "NIL\nNIL");
+make_test ("(hash-table-count tbl)", "1");
+make_test ("(clrhash tbl)", "#<HASH-TABLE EQUAL 0/1024>");
+make_test ("(hash-table-count tbl)", "0");
 make_test ("(last '(1 2 3))", "(3)");
 make_test ("(last '(1 2 3) 0)", "NIL");
 make_test ("(last '(1 2 3) 2)", "(2 3)");
@@ -695,7 +799,7 @@ make_test ("(read-preserving-whitespace (make-string-input-stream \"hello\"))",
 make_test ("(read-preserving-whitespace (make-string-input-stream \"123\"))", "123");
 make_test ("(eval '(write \"\"))", "\"\"\n\"\"");
 make_test ("(setq var 10)", "10");
-make_test ("(let ((var 12)) (eval 'var))", "10");
+make_test ("(let ((var 12)) (eval 'var))", "12");
 make_test ("(write \"\")", "\"\"\n\"\"");
 make_test ("(write-string \"aaa\\n\")", "aaan\n\"aaan\"");
 make_test ("(write-string \"\n" .
@@ -707,6 +811,9 @@ make_test ("(fresh-line)", "NIL");
 make_test ("(terpri)", "\nNIL");
 make_test ("(write-line \"aaa\")", "aaa\n\"aaa\"");
 make_test ("(write-sequence \"hello world\" *standard-output* :start 1 :end 5)", "ello\n\"hello world\"");
+make_test ("(let ((*print-base* 8)) (write 8))", "10\n8");
+make_test ("(let ((*print-base* 16)) (write 32))", "20\n32");
+make_test ("(let ((*print-base* 16)) (write 32/19))", "20/13\n32/19");
 make_test ("(prin1 '|Aa,a|)", "|Aa,a|\n|Aa,a|");
 make_test ("(prin1 #\\a)", "#\\a\n#\\a");
 make_test ("(prin1 \"aaa\" *standard-output*)", "\"aaa\"\n\"aaa\"");
@@ -715,7 +822,18 @@ make_test ("(princ #\\a)", "a\n#\\a");
 make_test ("(princ \"aaa\" *standard-output*)", "aaa\n\"aaa\"");
 make_test ("(print \"aaa\" *standard-output*)", "\n\"aaa\" \n\"aaa\"");
 make_test ("(let ((c 0)) (loop (if (= c 2) (return 0)) (incf c)))", "0");
+make_test ("(loop with i = (+ 0) with j = 1 do (incf i) (decf j) (if (= i 5) (return (list i j))))", "(5 -4)");
+make_test ("(loop with i = (+ 0) with j = (+ i 1) do (incf i) (decf j) (if (= i 5) (return (list i j))))", "(5 -4)");
+make_test ("(let ((i 0)) (loop with i = 1 and j = (+ i 1) do (incf i) (decf j) (if (= i 5) (return (list i j)))))", "(5 -3)");
+make_test ("(loop with i = 1 return i)", "1");
+make_test ("(loop with i = 0 and j = 5 do (incf i) (incf j) if (> i 1) if (> j 7) do (write 'done) (return (list i j)))", "DONE\n(3 8)");
+make_test ("(loop with i = 0 do (incf i) unless (> i 5) do (write i) else do (return nil))", "12345\nNIL");
+make_test ("(loop named foo for i from 0 if (= i 5) return i)", "5");
+make_test ("(loop :when nil :do 't :else :when nil :do 't :else :return 't)", "T");
+make_test ("(loop for i in '(0 1 2 3 4) if (numberp i) if (oddp i) collect i into l end finally (return l))", "(1 3)");
+make_test ("(let ((i 0)) (loop when (= i 0) do (write 'stop) :and do (return nil)))", "STOP\nNIL");
 make_test ("(loop initially (write 'hello) for i from 1 to 5 finally (write 'world) do (write i))", "HELLO12345WORLD\nNIL");
+make_test ("(loop for i to 2 do (write i) initially (write i))", "0012\nNIL");
 make_test ("(loop for i from 1 to 5 do (write i))", "12345\nNIL");
 make_test ("(loop for i from 1 below 5 do (write i))", "1234\nNIL");
 make_test ("(loop for i from 10 downto 5 by 2 do (write i))", "1086\nNIL");
@@ -723,6 +841,25 @@ make_test ("(loop for i upfrom 5 to 10 do (write i))", "5678910\nNIL");
 make_test ("(loop for i from 5 upto 11 for j downfrom 20 above 15 do (write i) (write j))", "520619718817916\nNIL");
 make_test ("(loop named foo for i from 0 do (write i) (if (> i 4) (return-from foo 0)))", "012345\n0");
 make_test ("(loop as i from 0 do (if (> i 5) (return 10) (write i)) finally (write 'finished))", "012345\n10");
+make_test ("(loop for i in '(0 1 2) do (write i))", "012\nNIL");
+make_test ("(loop for (i . j) in '((0 . 1) (1 . 2)) do (write j))", "12\nNIL");
+make_test ("(loop for (i . j) in nil finally (write 'done))", "DONE\nNIL");
+make_test ("(loop for i in '(0 1 2 3) by #'cddr do (write i))", "02\nNIL");
+make_test ("(loop for i from 10 downto 2 collect i)", "(10 9 8 7 6 5 4 3 2)");
+make_test ("(loop :for i :in '(0 1 2) :when (= i 2) :collect i :into out :and :collect i)", "(2)");
+make_test ("(loop for i in '((0 1) (2 3)) append i into l finally (return l))", "(0 1 2 3)");
+make_test ("(loop for i on nil do (write i))", "NIL");
+make_test ("(loop for i on '(0 1 2 3) do (write i))", "(0 1 2 3)(1 2 3)(2 3)(3)\nNIL");
+make_test ("(loop for i on '(0 1 2 3 4) by #'cddr do (write i))", "(0 1 2 3 4)(2 3 4)(4)\nNIL");
+make_test ("(loop for i = 0 then (incf i) do (write i) if (= i 4) return 10)", "01234\n10");
+make_test ("(let ((j 0)) (loop as i = (incf j) do (write i) if (= i 5) return 11))", "12345\n11");
+make_test ("(defparameter tbl (make-hash-table))", "TBL");
+make_test ("(setf (gethash 1 tbl) 5)", "5");
+make_test ("(setf (gethash 2 tbl) 6)", "6");
+make_test ("(setf (gethash 3 tbl) 7)", "7");
+make_test ("(loop for k being each hash-key in tbl do (write k))", "321\nNIL");
+make_test ("(loop for k being each hash-key in (make-hash-table) do (write k))", "NIL");
+make_test ("(loop for v being the hash-values in tbl using (hash-key k) do (write v) (write k))", "736251\nNIL");
 make_test ("(format t \"aaa\")", "aaa\nNIL");
 make_test ("(format t \"aaa~~\")", "aaa~\nNIL");
 make_test ("(format t \"aaa~%~%\")", "aaa\n\nNIL");
@@ -734,7 +871,7 @@ make_test ("(write \"hello\" :stream str)", "\"hello\"");
 make_test ("(write #\\a :stream str :stream *standard-input*)", "#\\a");
 make_test ("(write 1 :stream str)", "1");
 make_test ("(write 1/2 :stream str)", "1/2");
-make_test ("(write 0.1 :stream str)", "0.1");
+make_test ("(write 0.1 :stream str)", "0.0999999999999999916733");
 make_test ("(write (complex 1 1) :stream str)", "#C(1 1)");
 make_test ("(write-char #\\newline str)", "#\\Newline");
 make_test ("(write '(1 2 3) :stream str)", "(1 2 3)");
@@ -747,14 +884,14 @@ make_test ("(write \"bbb\" :stream str)", "\"bbb\"");
 make_test ("(write 'ccc :stream str)", "CCC");
 make_test ("(write #\\c :stream str)", "#\\c");
 make_test ("(write 100 :stream str)", "100");
-make_test ("(write 1.2 :stream str)", "1.2");
-make_test ("(get-output-stream-string str)", "\"AAA\\\"bbb\\\"CCC#\\\\c1001.2\"");
+make_test ("(write 1.2 :stream str)", "1.19999999999999995559");
+make_test ("(get-output-stream-string str)", "\"AAA\\\"bbb\\\"CCC#\\\\c1001.19999999999999995559\"");
 make_test ("(get-output-stream-string str)", "\"\"");
-make_test ("(gensym)", "#:G21");
-make_test ("(gensym)", "#:G22");
+make_test ("(gensym)", "#:G178");
+make_test ("(gensym)", "#:G179");
 make_test ("(gensym 5)", "#:G5");
-make_test ("(gensym \"A\")", "#:A23");
-make_test ("(gensym)", "#:G24");
+make_test ("(gensym \"A\")", "#:A180");
+make_test ("(gensym)", "#:G181");
 make_test ("(gentemp)", "T1");
 make_test ("(gentemp \"S\")", "S2");
 make_test ("'s3", "S3");
@@ -781,13 +918,19 @@ make_test ("(the integer (values 0 1))", "0\n1");
 make_test ("(prog1 2 (write \"hello\"))", "\"hello\"\n2");
 make_test ("(prog2 2 3 (write \"hello\"))", "\"hello\"\n3");
 make_test ("(destructuring-bind ((x)) '((0)) x)", "0");
+make_test ("(let ((z 0)) (destructuring-bind (i) (list z) (setq z 0)))", "0");
+make_test ("(let ((z 0)) (destructuring-bind (i) (list z) (setq z (+ i 1))))", "1");
 make_test ("(destructuring-bind (x (y (z))) '(0 (1 (2))) (list x y z))", "(0 1 2)");
 make_test ("(destructuring-bind (x (y z) w) '(0 ((1 2) 3) 4) (list x y z w))", "(0 (1 2) 3 4)");
+make_test ("(destructuring-bind (x . y) '(0 1 2 3) (list x y))", "(0 (1 2 3))");
 make_test ("(defstruct ship x y)", "SHIP");
 make_test ("#'make-ship", "#<FUNCTION MAKE-SHIP>");
 make_test ("#'ship-x", "#<FUNCTION SHIP-X>");
 make_test ("(defparameter s1 (make-ship))", "S1");
 make_test ("s1", "#<STRUCTURE OF CLASS SHIP>");
+make_test ("(class-of s1)", "#<STRUCTURE CLASS SHIP>");
+make_test ("(class-name (class-of s1))", "SHIP");
+make_test ("(type-of s1)", "SHIP");
 make_test ("(typep s1 'ship)", "T");
 make_test ("(typep 0 'ship)", "NIL");
 make_test ("(ship-x s1)", "NIL");
@@ -796,22 +939,54 @@ make_test ("(setf (ship-x s1) 0)", "0");
 make_test ("(setf (ship-y s1) 1)", "1");
 make_test ("(ship-x s1)", "0");
 make_test ("(ship-y s1)", "1");
+make_test ("(setf (ship-x s1) 2 (ship-y s1) 3)", "3");
 make_test ("(defclass starship () (x y speed))", "#<STANDARD-CLASS STARSHIP>");
 make_test ("(find-class 'starship)", "#<STANDARD-CLASS STARSHIP>");
 make_test ("(defparameter inst (make-instance 'starship))", "INST");
 make_test ("inst", "#<STARSHIP OBJECT ...>");
 make_test ("(class-of inst)", "#<STANDARD-CLASS STARSHIP>");
+make_test ("(class-name (class-of inst))", "STARSHIP");
+make_test ("(type-of inst)", "STARSHIP");
+make_test ("(typep inst 'starship)", "T");
+make_test ("(typep inst 'standard-object)", "T");
+make_test ("(typep 0 'starship)", "NIL");
 make_test ("(setf (slot-value inst 'x) 0)", "0");
 make_test ("(setf (slot-value inst 'y) 2.5)", "2.5");
 make_test ("(slot-value inst 'x)", "0");
 make_test ("(slot-value inst 'y)", "2.5");
+make_test ("(defclass animal nil (leg-number))", "#<STANDARD-CLASS ANIMAL>");
+make_test ("(defclass dog (animal) (hair-length))", "#<STANDARD-CLASS DOG>");
+make_test ("(defparameter d (make-instance 'dog))", "D");
+make_test ("(setf (slot-value d 'leg-number) 4)", "4");
+make_test ("(slot-value d 'leg-number)", "4");
+make_test ("(setf (slot-value d 'hair-length) 'short)", "SHORT");
+make_test ("(slot-value d 'hair-length)", "SHORT");
 make_test ("(defgeneric genfun (x y z))", "#<STANDARD-GENERIC-FUNCTION GENFUN>");
 make_test ("(defmethod genfun ((x string) y z) 10)", "#<STANDARD-METHOD GENFUN>");
 make_test ("(defmethod genfun ((x integer) y z) 11)", "#<STANDARD-METHOD GENFUN>");
 make_test ("(genfun \"\" nil nil)", "10");
 make_test ("(genfun 0 nil nil)", "11");
+make_test ("(defmethod genfun2 ((x string) y z) 12)", "#<STANDARD-METHOD GENFUN2>");
+make_test ("(defmethod genfun2 ((x integer) y z) 13)", "#<STANDARD-METHOD GENFUN2>");
 make_test ("(define-condition new-math-error (arithmetic-error) (a b))", "NEW-MATH-ERROR");
 make_test ("(make-condition 'new-math-error)", "#<CONDITION OF CLASS NEW-MATH-ERROR>");
+make_test ("(class-of (make-condition 'new-math-error))", "#<CONDITION-CLASS NEW-MATH-ERROR>");
+make_test ("(class-name (class-of (make-condition 'new-math-error)))", "NEW-MATH-ERROR");
+make_test ("(type-of (make-condition 'new-math-error))", "NEW-MATH-ERROR");
+make_test ("(typep (make-condition 'new-math-error) 'new-math-error)", "T");
+make_test ("(typep (make-condition 'new-math-error) 'arithmetic-error)", "T");
+make_test ("(typep (make-condition 'new-math-error) 'file-error)", "NIL");
+make_test ("(typep 0 'new-math-error)", "NIL");
+make_test ("(define-condition strange-condition nil nil)", "STRANGE-CONDITION");
+make_test ("(typep (make-condition 'strange-condition) 'condition)", "T");
+make_test ("(simple-condition-format-control (make-condition 'simple-condition))", "NIL");
+make_test ("(arithmetic-error-operation (make-condition 'arithmetic-error))", "NIL");
+make_test ("(declaim (optimize speed (safety 0) (debug 3)))", "T");
+make_test ("(declaim (ignorable a) (ignore b c) (inline e) (notinline d f))", "T");
+make_test ("(declaim (type list a) (ftype (function nil nil) b))", "T");
+make_test ("(proclaim '(optimize speed (safety 0) (debug 3)))", "T");
+make_test ("(proclaim '(ignorable a))", "T");
+make_test ("(proclaim '(type list a))", "T");
 make_test ("(tagbody\n" .
 	   "  (write \"1\")\n" .
 	   "  (go jmp)\n" .
@@ -821,6 +996,7 @@ make_test ("(tagbody\n" .
 make_test ("(tagbody 1 (go 3) 2 3 (go 4) 4 (write \"\"))", "\"\"\nNIL");
 make_test ("(tagbody 1 2 (tagbody 3 (go 4)) 4 (write \"\"))", "\"\"\nNIL");
 make_test ("(tagbody (write 1) (block nil (write 2) (go c)) c (write 4))", "124\nNIL");
+make_test ("(tagbody (go c) c)", "NIL");
 make_test ("(block test (write 1) (write 2) (return-from test 10) (write 3))", "12\n10");
 make_test ("(block test (write 1) (block test2 (write 2) (return-from test 10) (write 3)))", "12\n10");
 make_test ("(block test (write 1) (block test2 (write 2) (return-from test (values 10 11)) (write 3)))", "12\n10\n11");
@@ -838,6 +1014,14 @@ make_test ("(restart-bind ((rest (lambda () (write 10)))) (invoke-restart 'rest)
 make_test ("(block test (unwind-protect 0 (write 'whatever)))", "WHATEVER\n0");
 make_test ("(block test (unwind-protect (return-from test 0) (write 'whatever)))", "WHATEVER\n0");
 make_test ("(block test (unwind-protect (return-from test (values 0 1)) (write 'whatever)))", "WHATEVER\n0\n1");
+make_test ("(signal \"blah\")", "NIL");
+make_test ("(handler-bind ((simple-condition (lambda (e) (write 'oh)))) (signal \"blah\"))", "OH\nNIL");
+make_test ("(handler-bind ((condition (lambda (e) (write 'ah))) (simple-condition (lambda (e) (write 'oh)))) (signal \"blah\"))", "AHOH\nNIL");
+make_test ("(handler-bind ((condition (lambda (e) (write (signal \"w\"))))) (signal \"blah\"))", "NIL\nNIL");
+make_test ("(block nil (handler-bind ((condition (lambda (e) (return 0)))) (signal \"blah\")))", "0");
+make_test ("(handler-bind ((condition (lambda (e) (write (simple-condition-format-control e))))) (signal \"blah\"))", "\"blah\"\nNIL");
+make_test ("(handler-bind ((arithmetic-error (lambda (e) (write 'oh)))) (signal 'arithmetic-error))", "OH\nNIL");
+make_test ("(error \"bah\")", "\"bah\"");
 make_test ("(cddr '(0 1 2))", "(2)");
 make_test ("(cddddr '(0 1 2 3 4))", "(4)");
 make_test ("(cadddr '(0 1 2 3 4))", "3");
@@ -869,6 +1053,9 @@ make_test ("(or (values 1 2) nil)", "1");
 make_test ("(concatenate 'string)", "\"\"");
 make_test ("(concatenate 'string \"\")", "\"\"");
 make_test ("(concatenate 'string \"aa\" \"bb\")", "\"aabb\"");
+make_test ("(concatenate 'vector #(0 1 2) #(3) #(4 5))", "#(0 1 2 3 4 5)");
+make_test ("(concatenate 'list '(0 1 2) nil '(3 4 5))", "(0 1 2 3 4 5)");
+make_test ("(concatenate 'list nil)", "NIL");
 make_test ("(do (x (y 0) (z 0 (1+ z))) ((>= z 10) (values x y z)) (write z))", "0123456789\nNIL\n0\n10");
 make_test ("(do ((x 0 (1+ x)) (y 0 (1+ x))) ((>= y 10)) (write y))", "0123456789\nNIL");
 make_test ("(do ((dd 0 (1+ dd))) (nil) (if (> dd 5) (return 10) (write dd)))", "012345\n10");
@@ -884,12 +1071,25 @@ make_test ("(dolist (i '(0 1 2 3)) (write i))", "0123\nNIL");
 make_test ("(dolist (i '(0 1 2 3)) (if (= i 2) (return (values 12 13))) (write i))", "01\n12\n13");
 make_test ("(mapcar #'car '((1 a) (2 b) (3 c)))", "(1 2 3)");
 make_test ("(mapcar #'abs '(3 -4 2 -5 -6))", "(3 4 2 5 6)");
+make_test ("(mapcar 'abs '(3 -4 2 -5 -6))", "(3 4 2 5 6)");
 make_test ("(mapcar #'cons '(a b c) '(1 2 3))", "((A . 1) (B . 2) (C . 3))");
 make_test ("(let ((s 0)) (write (mapcar (lambda (s) s) '(1 2 3))) s)", "(1 2 3)\n0");
+make_test ("(map 'string #'char-upcase \"abcDefGh\")", "\"ABCDEFGH\"");
+make_test ("(map 'vector #'+ #(1 2 3 4) '(10 11 12) #(20 21 22))", "#(31 34 37)");
+make_test ("(map 'list #'+ #(1 2 3 4) '(10 11 12) #(20 21 22))", "(31 34 37)");
+make_test ("(map nil #'+ '(1 2 3 4) '(10 11 12) '(20 21 22))", "NIL");
+make_test ("(map 'vector #'+ #(1 2 3 4) nil #(20 21 22))", "#()");
 make_test ("(remove 3 '(1 2 3 4 3 2 1))", "(1 2 4 2 1)");
 make_test ("(remove #\\e \"abcdefg\")", "\"abcdfg\"");
 make_test ("(remove-if #'numberp #(1 2 3))", "#()");
+make_test ("(remove-if 'numberp #(1 2 3))", "#()");
+make_test ("(remove-if #'not '(nil))", "NIL");
 make_test ("(remove-if-not #'listp #(()))", "#(NIL)");
+make_test ("(remove-duplicates \"abcaadeabfc\")", "\"deabfc\"");
+make_test ("(remove-duplicates '(0 1 2 3 4 3 3 5 6 8 6))", "(0 1 2 4 3 5 8 6)");
+make_test ("(remove-duplicates nil)", "NIL");
+make_test ("(remove-duplicates #(0 1 2 4 5 3 4 6 9 3) :start 1 :end 7 :test #'=)", "#(0 1 2 5 3 4 6 9 3)");
+make_test ("(delete-duplicates #(0 1 2 4 5 3 4 6 9 3) :from-end t :test #'=)", "#(0 1 2 4 5 3 6 9)");
 make_test ("(reverse '(0 1 2 3))", "(3 2 1 0)");
 make_test ("(reverse \"hello\")", "\"olleh\"");
 make_test ("(reverse #(0 1 2 3))", "#(3 2 1 0)");
@@ -905,6 +1105,8 @@ make_test ("(push (1+ 0) ll)", "(1)");
 make_test ("ll", "(1)");
 make_test ("(pop ll)", "1");
 make_test ("ll", "NIL");
+make_test ("(set-difference '(0 1 1 2) nil)", "(0 1 1 2)");
+make_test ("(nset-difference '(0 1 1 2) '(3 1 5) :test #'=)", "(0 2)");
 make_test ("(cond (t (+ 1 2)))", "3");
 make_test ("(cond (nil 2) (t 3) (nil 4))", "3");
 make_test ("(cond ((= 2 (+ 1 2)) (+ 0 10)) ((= 3 (+ 1 2)) (+ 12) (+ 0 11)))", "11");
@@ -916,6 +1118,9 @@ make_test ("(case (+ 1 2) (2 10) ((4 5) (+ 1 5)) (otherwise 11 12))", "12");
 make_test ("(multiple-value-bind (x y) (floor 2 1) (list x y))", "(2 0)");
 make_test ("(multiple-value-bind (x) (floor 2 1) (list x))", "(2)");
 make_test ("(multiple-value-bind (x y z) (floor 2 1) (list x y z))", "(2 0 NIL)");
+make_test ("(multiple-value-bind (x y) (values 'i 'j) (list x y))", "(I J)");
+make_test ("(prog ((x 0) y) a (write 0) (go c) b (write 1) c (return x))", "0\n0");
+make_test ("(prog* ((x 0) (y x)) a (write 0) (go c) b (write 1) c (return y))", "0\n0");
 make_test ("(every #'evenp '(0 2 4))", "T");
 make_test ("(every #'< '(0 1 2) '(1 1))", "NIL");
 make_test ("(some #'< '(0 1 2) '(1 1))", "T");
@@ -970,6 +1175,7 @@ make_test ("#+common-lisp '(1\n2)", "(1 2)");
 make_test ("#-common-lisp (\n) 0", "0");
 make_test ("#+foo \"aaa\nbbb\" 0", "0");
 make_test ("#+foo (1\n2 (3)) 0", "0");
+make_test ("#-fooo :fooo 0", ":FOOO\n0");
 
 make_test ("*package*", "#<PACKAGE \"COMMON-LISP-USER\">");
 make_test ("(find-package *package*)", "#<PACKAGE \"COMMON-LISP-USER\">");
@@ -997,7 +1203,11 @@ make_test ("(in-package cl-user)", "#<PACKAGE \"COMMON-LISP-USER\">");
 make_test ("*package*", "#<PACKAGE \"COMMON-LISP-USER\">");
 make_test ("(use-package 'cl)", "T");
 make_test ("(make-package \"newtest\")", "#<PACKAGE \"newtest\">");
+make_test ("(rename-package (find-package '|newtest|) (find-package '|newtest|))", "#<PACKAGE \"newtest\">");
 make_test ("(in-package \"newtest\")", "#<PACKAGE \"newtest\">");
+make_test ("'a", "A");
+make_test ("'b", "B");
+make_test ("(cl:do-symbols (s '|newtest|) (cl:write s))", "ABS|newtest|\nCOMMON-LISP:NIL");
 make_test ("(cl:package-use-list cl:*package*)", "COMMON-LISP:NIL");
 make_test ("(cl:use-package 'cl)", "T");
 make_test ("(cl:use-package nil \"newtest\")", "T");
@@ -1010,6 +1220,7 @@ make_test ("(cl:export 'cdr)", "T");
 make_test ("'|newtest|:cdr", "CDR");
 make_test ("(cl:export 'jjj)", "T");
 make_test ("(cl:export '(jjj) cl:*package*)", "T");
+make_test ("(do-external-symbols (s '|newtest|) (write s))", "CDRJJJ\nNIL");
 make_test ("'|newtest|:jjj", "JJJ");
 make_test ("(cl:unexport nil \"newtest\")", "T");
 make_test ("(cl:unexport '(jjj) cl:*package*)", "T");
@@ -1023,6 +1234,8 @@ make_test ("(defpackage \"aaa\" (:nicknames \"bbb\" |ccc|) (:use cl-user) (:expo
 make_test ("(package-nicknames \"aaa\")", "(\"bbb\" \"ccc\")");
 make_test ("(package-use-list \"aaa\")", "(#<PACKAGE \"COMMON-LISP-USER\">)");
 make_test ("'|aaa|:|sym|", "|aaa|:|sym|");
+make_test ("(make-package 'whatever :nicknames '(\"APPLE\" pear))", "#<PACKAGE \"WHATEVER\">");
+make_test ("(package-nicknames 'whatever)", "(\"APPLE\" \"PEAR\")");
 make_test ("(defpackage \"aaa\" (:nicknames \"bbb\" |ccc|) (:use cl-user) (:export \"sym\") (:intern \"eee\" \"fff\") (:import-from \"CL\" car caar) (:import-from \"CL-USER\" al-argc))", "#<PACKAGE \"aaa\">");
 
 
@@ -1031,24 +1244,38 @@ make_test ("(defpackage \"aaa\" (:nicknames \"bbb\" |ccc|) (:use cl-user) (:expo
 make_test ("2/4", "1/2");
 make_test ("0.1s2", "10.0");
 make_test ("0.0", "0.0");
-make_test ("0.1", "0.1");
+make_test ("0.1", "0.0999999999999999916733");
 make_test ("(+)", "0");
 make_test ("(+ 1)", "1");
 make_test ("(+ 1 2 3)", "6");
 make_test ("(+ 1/2 2 3 1000)", "2011/2");
 make_test ("(+ a b)", "17");
-make_test ("(+ .1 .1 .1)", "0.3");
+make_test ("(+ .1 .1 .1)", "0.299999999999999988898");
+make_test ("(+ 1 1 1.2)", "3.20000000000000017764");
+make_test ("(+ #c(1 1) #c(2.0 3))", "#C(3.0 4.0)");
+make_test ("(+ #c(1 1) #c(2 -1))", "3");
+make_test ("(+ 1 #c(2 3.0))", "#C(3.0 3.0)");
 make_test ("(- 3.5)", "-3.5");
 make_test ("(- 3 4.5)", "-1.5");
+make_test ("(- 1 1 1.2)", "-1.19999999999999995559");
+make_test ("(- #c(1 2) #c(3.0 4))", "#C(-2.0 -2.0)");
+make_test ("(- 1 #c(3.0 4))", "#C(-2.0 -4.0)");
 make_test ("(*)", "1");
 make_test ("(* 1)", "1");
 make_test ("(* 3 5)", "15");
+make_test ("(* 1 1 1.2)", "1.19999999999999995559");
+make_test ("(* 3 #c(3.0 4))", "#C(9.0 12.0)");
+make_test ("(* #c(0 1) #c(0 1))", "-1");
 make_test ("(/ 12 4)", "3");
 make_test ("(/ 13 4)", "13/4");
 make_test ("(/ -8)", "-1/8");
 make_test ("(/ 3 4 5)", "3/20");
 make_test ("(/ 0.5)", "2.0");
 make_test ("(/ 60 -2 3 5.0)", "-2.0");
+make_test ("(/ #c(1 1) #c(1 1))", "1");
+make_test ("(/ #c(1 2) #c(1 1))", "#C(3/2 1/2)");
+make_test ("(/ #c(1 2) 3)", "#C(1/3 2/3)");
+make_test ("(/ #c(1.1 2) 3)", "#C(0.36666666666666658525 0.666666666666666629659)");
 make_test ("(floor 1)", "1\n0");
 make_test ("(floor 1/2)", "0\n1/2");
 make_test ("(ffloor 1)", "1.0\n0");
@@ -1059,10 +1286,10 @@ make_test ("(fceiling 2)", "2.0\n0");
 make_test ("(fceiling 1/2 1)", "1.0\n-1/2");
 make_test ("(truncate -3 2)", "-1\n-1");
 make_test ("(truncate -1)", "-1\n0");
-make_test ("(truncate -.1)", "0\n-0.1");
+make_test ("(truncate -.1)", "0\n-0.0999999999999999916733");
 make_test ("(ftruncate -3 2)", "-1.0\n-1");
 make_test ("(ftruncate -1)", "-1.0\n0");
-make_test ("(ftruncate -.1)", "0.0\n-0.1");
+make_test ("(ftruncate -.1)", "0.0\n-0.0999999999999999916733");
 make_test ("(round 5)", "5\n0");
 make_test ("(round .5)", "0\n0.5");
 make_test ("(round -1.5)", "-2\n0.5");
@@ -1073,6 +1300,15 @@ make_test ("(numerator 2)", "2");
 make_test ("(numerator 2/3)", "2");
 make_test ("(denominator 2)", "1");
 make_test ("(denominator 2/3)", "3");
+make_test ("(gcd)", "0");
+make_test ("(gcd -1)", "1");
+make_test ("(gcd 0 -3)", "3");
+make_test ("(gcd -4 6)", "2");
+make_test ("(gcd 6 -27 81)", "3");
+make_test ("(lcm)", "1");
+make_test ("(lcm -1)", "1");
+make_test ("(lcm 4 -6)", "12");
+make_test ("(lcm -9 3 27)", "27");
 make_test ("(sqrt 9)", "3.0");
 make_test ("(sqrt 9.0)", "3.0");
 make_test ("(isqrt 10)", "3");
@@ -1087,6 +1323,7 @@ make_test ("#c (0 1)", "#C(0 1)");
 make_test ("#C(1 1.0)", "#C(1.0 1.0)");
 make_test ("(realpart 1)", "1");
 make_test ("(realpart (complex 1 2.0))", "1.0");
+make_test ("(typep (realpart #c(1 1/2)) 'integer)", "T");
 make_test ("(imagpart 1)", "0");
 make_test ("(imagpart (complex 1 1/2))", "1/2");
 make_test ("(conjugate 1)", "1");
@@ -1097,9 +1334,13 @@ make_test ("(= 1 1)", "T");
 make_test ("(= 1 2)", "NIL");
 make_test ("(= 1 1 1.0 1/1)", "T");
 make_test ("(= 0 -0.0)", "T");
+make_test ("(= #c(0 1) #c(0.0 1))", "T");
+make_test ("(= #c(2 0) 1)", "NIL");
+make_test ("(= #c(1.0 0) 1)", "T");
 make_test ("(/= 0)", "T");
 make_test ("(/= 0 1 2.0)", "T");
 make_test ("(/= 0 1.0 2 1)", "NIL");
+make_test ("(/= 1 2 #c(1 0) 3)", "NIL");
 make_test ("(< 0)", "T");
 make_test ("(< 0 1.0 2)", "T");
 make_test ("(< 0 0.0 1)", "NIL");
@@ -1110,7 +1351,7 @@ make_test ("(> 1 1)", "NIL");
 make_test ("(>= 1 1.0)", "T");
 make_test ("(min 0)", "0");
 make_test ("(min 0 1 2)", "0");
-make_test ("(min 0 1/2 -0.1)", "-0.1");
+make_test ("(min 0 1/2 -0.1)", "-0.0999999999999999916733");
 make_test ("(max 0)", "0");
 make_test ("(max 0 1.0 2)", "2");
 make_test ("(max 0 1/2 -0.1)", "1/2");
@@ -1186,6 +1427,7 @@ make_test ("(oddp 0)", "NIL");
 make_test ("(oddp -1)", "T");
 make_test ("*random-state*", "#<RANDOM-STATE ?>");
 make_test ("(make-random-state)", "#<RANDOM-STATE ?>");
+make_test ("(type-of (make-random-state))", "RANDOM-STATE");
 make_test ("(make-random-state t)", "#<RANDOM-STATE ?>");
 make_test ("(make-random-state nil)", "#<RANDOM-STATE ?>");
 make_test ("(make-random-state *random-state*)", "#<RANDOM-STATE ?>");
